import socket
import base64
import logging
import subprocess
from pathlib import Path

from pyroute2 import IPDB, WireGuard, CreateException
from nacl.public import PrivateKey

from platform_agent.cmd.lsmod import module_loaded
from platform_agent.cmd.wg_show import get_wg_listen_port
from platform_agent.routes import ip_route_add, ip_route_del

logger = logging.getLogger()


class WgConfException(Exception):
    pass


class WgConf():

    def __init__(self):

        self.wg_kernel = module_loaded('wireguard')
        self.wg = WireGuard() if self.wg_kernel else WireguardGo()

    def get_wg_keys(self, ifname):
        private_key_path = f"/etc/noia-agent/privatekey-{ifname}"
        public_key_path = f"/etc/noia-agent/publickey-{ifname}"
        private_key = Path(private_key_path)
        public_key = Path(public_key_path)
        if not private_key.is_file() or not public_key.is_file():
            privKey = PrivateKey.generate()
            pubKey = base64.b64encode(bytes(privKey.public_key))
            privKey = base64.b64encode(bytes(privKey))
            base64_privKey = privKey.decode('ascii')
            base64_pubKey = pubKey.decode('ascii')
            private_key.write_text(base64_privKey)
            public_key.write_text(base64_pubKey)
            private_key.chmod(0o600)
            public_key.chmod(0o600)

        if self.wg_kernel:
            return public_key.read_text().strip(), private_key.read_text().strip()
        else:
            return public_key.read_text().strip(), private_key_path


    def next_free_port(self, port=1024, max_port=65535):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while port <= max_port:
            try:
                sock.bind(('', port))
                sock.close()
                return port
            except OSError:
                port += 1
        raise IOError('no free ports')

    def create_interface(self, ifname, internal_ip, listen_port=None):
        logger.info(f"[WG_CONF] - Creating interface {ifname}")
        public_key, private_key = self.get_wg_keys(ifname)

        with IPDB() as ip:
            if self.wg_kernel:
                try:
                    wg_int = ip.create(kind='wireguard', ifname=ifname)
                except CreateException as e:
                    raise WgConfException(str(e))
            else:
                self.wg.create_interface(ifname)
                if ifname in ip.interfaces:
                    wg_int = ip.interfaces[ifname]
                else:
                    raise WgConfException("Wireguard-go failed to create interface")
            wg_int.add_ip(internal_ip)
            wg_int.up()
            wg_int.commit()
        self.wg.set(
            ifname,
            private_key=private_key,
            listen_port=listen_port
        )

        listen_port = self.get_listening_port(ifname)

        return {
            "public_key": public_key,
            "listen_port": listen_port
        }

    def add_peer(self, ifname, public_key, allowed_ips, gw_ipv4, endpoint_ipv4=None, endpoint_port=None):
        peer = {'public_key': public_key,
                'endpoint_addr': endpoint_ipv4,
                'endpoint_port': endpoint_port,
                'persistent_keepalive': 15,
                'allowed_ips': allowed_ips}
        self.wg.set(ifname, peer=peer)
        ip_route_add(ifname, allowed_ips, gw_ipv4)
        return

    def remove_peer(self, ifname, public_key, allowed_ips=None):
        peer = {
            'public_key': public_key,
            'remove': True
            }

        self.wg.set(ifname, peer=peer)
        if allowed_ips:
            ip_route_del(ifname, allowed_ips)
        return

    def remove_interface(self, ifname):
        with IPDB() as ipdb:
            if ifname not in ipdb.interfaces:
                raise WgConfException(f'[{ifname}] does not exist')
            with ipdb.interfaces[ifname] as i:
                i.remove()
        return

    def get_listening_port(self, ifname):
        if self.wg_kernel:
            wg_info = dict(self.wg.info(ifname)[0]['attrs'])
            return wg_info['WGDEVICE_A_LISTEN_PORT']

        else:
            wg_info = self.wg.info(ifname)
            return wg_info['listen_port']


class WireguardGo():

    def set(self, ifname, peer=None, private_key=None, listen_port=None):
        full_cmd = f"wg set {ifname}".split(' ')
        if peer:
            allowed_ips_cmd = ""
            if not peer.get('remove'):
                for ip in peer.get('allowed_ips', []):
                    allowed_ips_cmd += f"allowed-ips {ip} "
                peer_cmd = f"peer {peer['public_key']} {allowed_ips_cmd}endpoint {peer['endpoint_addr']}:{peer['endpoint_port']}".split(' ')
            else:
                peer_cmd = f"peer {peer['public_key']} remove"
            full_cmd += peer_cmd
        if private_key:
            private_key_cmd = f"private-key {private_key}".split(' ')
            full_cmd += private_key_cmd
        if listen_port:
            listen_port_cmd = f"listen-port {listen_port}".split(' ')
            full_cmd += listen_port_cmd

        result_set = subprocess.run(full_cmd, encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        complete_output = result_set.stdout or result_set.stderr
        complete_output = complete_output or 'Success'
        logger.info(f"[Wireguard-go] - WG SET - {complete_output} , args {full_cmd}")
        return complete_output

    def create_interface(self, ifname):
        try:
            result_set = subprocess.run(['wireguard-go', ifname], encoding='utf-8', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except FileNotFoundError:
            raise WgConfException(f'Wireguard-go missing')

        complete_output = result_set.stdout or result_set.stderr
        complete_output = complete_output or 'Success'
        return complete_output

    def info(self, ifname):
        return {
            "listen_port": get_wg_listen_port(ifname)
        }


from pytest import fixture


@fixture
def standard_coin_set():
    return [1, 2, 5, 10, 20, 50]


@fixture
def custom_coin_set():
    return [1, 3, 10, 22, 84]


@fixture
def custom_coin_set_even():
    return [2, 4, 8, 16]


@fixture
def list_of_coins():
    return [7, 7, 7, 5, 4, 3, 4, 6, 7, 7, 99]



#!/usr/bin/env python3
"""
Noia platform agent.
"""

__author__ = "NOIA Network"
__version__ = "0.0.41"
__license__ = "MIT"

import os
import argparse

from platform_agent.config.logger import configure_logger
from platform_agent.config.settings import Config
from platform_agent.agent_websocket import WebSocketClient


def main(args=None):
    """ This is executed when run from the command line """
    parser = argparse.ArgumentParser()

    # Required positional argument
    parser.add_argument("run", help="Required parameter to start agent")

    # Optional verbosity counter (eg. -v, -vv, -vvv, etc.)
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Verbosity (-v, -vv, etc)")

    # Specify output of "--version"
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s (version {version})".format(version=__version__))

    args = parser.parse_args()
    agent(args)


def agent(args):
    """ Main entry point of the app """
    if args.run:

        # Retrieving settings from config on start
        Config()

        # Configuring logger globally
        configure_logger()

        # Initiating WS client
        client = WebSocketClient(
            os.environ.get('NOIA_CONTROLLER_URL', 'app-controller-platform-agents.noia.network'),
            os.environ['NOIA_API_KEY']
        )

        # Starting WS client main thread
        client.start()


if __name__ == "__main__":
    """ This is executed when run from the command line """
    main()


import logging
import threading
import os

from platform_agent.lib.get_info import gather_initial_info
from platform_agent.wireguard import WgConfException, WgConf, WireguardPeerWatcher
from platform_agent.docker_api.docker_api import DockerNetworkWatcher
from platform_agent.executors.wg_exec import WgExecutor

logger = logging.getLogger()


class AgentApi:

    def __init__(self, runner):
        self.runner = runner
        self.wg_peers = None
        self.wgconf = WgConf()
        self.wg_executor = WgExecutor(self.runner)
        threading.Thread(target=self.wg_executor.run).start()
        if os.environ.get("NOIA_NETWORK_API", '').lower() == "docker":
            self.network_watcher = DockerNetworkWatcher(self.runner).start()
        # self.rerouting = Rerouting().start()

    def call(self, type, data, request_id):
        result = None
        try:
            if hasattr(self, type):
                logger.info(f"[AGENT_API] Calling agent api {data}")
                if not isinstance(data, (dict, list)):
                    logger.error('[AGENT_API] data should be "DICT" type')
                    result = {'error': "BAD REQUEST"}
                else:
                    fn = getattr(self, type)
                    result = fn(data, request_id=request_id)
        except AttributeError as error:
            logger.warning(error)
            result = {'error': str(error)}
        return result

    def GET_INFO(self, data, **kwargs):
        return gather_initial_info(**data)

    def WG_INFO(self, data, **kwargs):
        if self.wg_peers:
            self.wg_peers.join(timeout=1)
            self.wg_peers = None
        self.wg_peers = WireguardPeerWatcher(self.runner, **data)
        self.wg_peers.start()
        logger.debug(f"[WIREGUARD_PEERS] Enabled | {data}")

    def WG_CONF(self, data, **kwargs):
        self.wg_executor.queue.put({"data": data, "request_id": kwargs['request_id']})
        return False

    def CONFIG_INFO(self, data, **kwargs):
        for vpn_cmd in data.get('vpn', []):
            try:
                fn = getattr(self.wgconf, vpn_cmd['fn'])
                fn(**vpn_cmd['args'])
            except WgConfException as e:
                logger.error(f"[CONFIG_INFO] {str(e)}")



import json
import logging
import threading
import time

from pyroute2 import WireGuard

from platform_agent.lib.ctime import now

logger = logging.getLogger()

IFNAME = 'wg0-server'


class WireguardPeerWatcher(threading.Thread):

    def __init__(self, client, interval):
        super().__init__()
        self.client = client
        self.interval = interval
        self.wg = WireGuard()
        self.stop_peer_watcher = threading.Event()
        self.daemon = True

    def run(self):
        while not self.stop_peer_watcher.is_set():
            results = []
            ss = self.wg.info(IFNAME)
            wg_info = dict(ss[0]['attrs'])
            peers = wg_info['WGDEVICE_A_PEERS']['attrs']
            for peer in peers:
                peer_dict = dict(peer[1]['attrs'])
                results.append(
                    {
                        "last_handshake": peer_dict['WGPEER_A_LAST_HANDSHAKE_TIME'],
                        "keep_alive_intreval": peer_dict['WGPEER_A_PERSISTENT_KEEPALIVE_INTERVAL'],
                        "public_key": peer_dict['WGPEER_A_PUBLIC_KEY'],
                        "rx_bytes": peer_dict['WGPEER_A_RX_BYTES'],
                        "tx_bytes": peer_dict['WGPEER_A_TX_BYTES'],

                    }
                )
            self.client.send(json.dumps({
                'id': "UNKNOWN",
                'executed_at': now(),
                'type': 'WIREGUARD_PEERS',
                'data': results
            }))
            time.sleep(int(self.interval))

    def join(self, timeout=None):
        self.stop_peer_watcher.set()
        super().join(timeout)



import subprocess


def module_loaded(module_name):
    """Checks if module is loaded"""
    lsmod_proc = subprocess.Popen(['lsmod'], stdout=subprocess.PIPE)
    grep_proc = subprocess.Popen(['grep', module_name], stdin=lsmod_proc.stdout)
    grep_proc.communicate()  # Block until finished
    return grep_proc.returncode == 0



import json
import logging
import threading
import time

import docker
from platform_agent.docker_api.helpers import format_networks_result
from platform_agent.lib.ctime import now

logger = logging.getLogger()


class DockerNetworkWatcher(threading.Thread):

    def __init__(self, ws_client):
        super().__init__()
        self.ws_client = ws_client
        self.docker_client = docker.from_env()
        self.events = self.docker_client.events(decode=True)
        self.daemon = True

    def run(self):
        for event in self.events:
            if event.get('Type') == 'network' and event.get('Action') in ['create', 'destroy']:
                networks = self.docker_client.networks()
                result = format_networks_result(networks)
                logger.info(f"[NETWORK_INFO] Sending networks {result}")
                self.ws_client.send(json.dumps({
                    'id': "ID." + str(time.time()),
                    'executed_at': now(),
                    'type': 'NETWORK_INFO',
                    'data': result
                }))

    def join(self, timeout=None):
        self.events.close()
        super().join(timeout)



import json
import logging
import threading
import time
import queue

from platform_agent.lib.ctime import now
from platform_agent.wireguard import WgConfException, WgConf
logger = logging.getLogger()


class WgExecutor(threading.Thread):

    CMD_TYPE = "WG_CONF"

    def __init__(self, client):
        super().__init__()
        self.queue = queue.Queue()
        self.client = client
        self.stop_wg_executor = threading.Event()
        self.daemon = True
        self.wg = None
        self.wgconf = WgConf()

        threading.Thread.__init__(self)

    def get_from_queue(self):
        payloads = {}
        t_end = time.time() + 1  # run for 1 second
        while time.time() < t_end:
            try:
                message = self.queue.get(timeout=0.1)
            except queue.Empty:
                continue
            request_id = message['request_id']
            payloads[request_id] = []
            data = message['data'] if type(message['data']) == list else [message['data']]
            for payload in data:
                try:
                    payloads[request_id].append(
                        {
                            "fn_name": payload['fn'],
                            "fn_args": payload['args'],
                            "request_id": request_id
                        }
                    )
                except AttributeError as e:
                    logger.warning(e)
                    payloads[request_id].append(
                        {"request_id": request_id, "error": f"{e}"})
        return payloads

    def run(self):

        while True:
            payloads = self.get_from_queue()
            if not payloads:
                continue
            logger.info(f"[WG_EXECUTOR] - Received {payloads}")
            for request_id in list(payloads.keys()):
                result = {}
                ok = {}
                errors = []
                for payload in payloads[request_id]:
                    if payload.get('error'):
                        errors.append(payload['error'])
                    else:
                        try:
                            fn = getattr(self.wgconf, payload['fn_name'])
                            if fn == self.wgconf.add_peer:
                                threading.Thread(target=self.add_peer, args=(payload, request_id)).run()
                                continue
                            ok.update({"fn": payload['fn_name'], "data": fn(**payload['fn_args']), "args": payload['fn_args']})
                        except WgConfException as e:
                            logger.error(f"[WG_EXECUTOR] failed. exception = {str(e)}, data = {payload}")
                            errors.append({payload['fn_name']: str(e), "args": payload['fn_args']})
                    if errors:
                        result['error'] = errors
                    elif ok:
                        result = ok
                    logger.info(f"[WG_EXECUTOR] - Results {result}")
                    self.client.send(json.dumps({
                        'id': request_id,
                        'executed_at': now(),
                        'type': self.CMD_TYPE,
                        'data': result
                    }))

    def add_peer(self, payload, request_id):
        result = {}
        try:
            result = self.wgconf.add_peer(**payload['fn_args'])
        except WgConfException as e:
            logger.error(f"[WG_EXECUTOR] failed. exception = {str(e)}, data = {payload}")
            result['error'] = {payload['fn_name']: str(e), "args": payload['fn_args']}
        self.client.send(json.dumps({
            'id': request_id,
            'executed_at': now(),
            'type': self.CMD_TYPE,
            'data': result
        }))

    def join(self, timeout=None):
        self.stop_wg_executor.set()
        super().join(timeout)



from setuptools import setup, find_packages, Extension


from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name="platform-agent-devel",
    version='0.0.6',
    py_modules=['platform-agent'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        'pyroute2==0.5.12',
        'websocket-client==0.57.0',
        'requests==2.22.0',
        'PyNaCl==1.3.0',
        'docker-py==1.10.6',
        'icmplibv2==1.0.5',
        'PyYAML==5.3.1'
    ],
    packages=find_packages(),
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    entry_points={
        'console_scripts': [
            'noia_agent = platform_agent.__main__:main'
        ]
    },
)
